use crate::client::{Client, ClientInfo};
use crate::database::{DatabaseSettings, DatabaseType, DatabaseCreationError};
use crate::response::MessageResponseData;
use crate::room::{Room, RoomSettings};

use std::collections::HashMap;
use std::rc::Rc;
use std::cell::RefCell;
use simple_websockets::{Event, Responder, Message, EventHub};
use derive_more::From;

#[derive(From, Debug)]
pub enum ServerCreationError {
    SimpleWebsocketsError(simple_websockets::Error),
    DatabaseCreationError(DatabaseCreationError),
}

#[derive(From, Debug)]
pub enum ServerError {
    DatabaseError(rusqlite::Error),
}

pub struct ServerSettings {
    pub port: u16,
    pub testing: bool,
}

impl Default for ServerSettings {
    fn default() -> Self {
        Self {
            port: 8080,
            testing: true,
        }
    }
}

pub struct Server {
    event_hub: EventHub,
    lobby: Rc<RefCell<Room>>,
    rooms: HashMap<String, Room>,
    clients: HashMap<u64, Client>,
}

impl Server {
    pub fn new(settings: &ServerSettings) -> Result<Self, ServerCreationError> {
        Ok(Self {
            event_hub: simple_websockets::launch(settings.port)?,
            lobby: Rc::new(RefCell::new(Room::new(RoomSettings {
                database_settings: DatabaseSettings {
                    db_type: match settings.testing {
                        true => DatabaseType::InMemory,
                        false => DatabaseType::OnDisk("shiritori.sb".to_string()),
                    },
                },
                ..Default::default()
            })?)),
            rooms: HashMap::new(),
            clients: HashMap::new(),
        })
    }

    pub fn get_client(&self, id: u64) -> Option<&Client> {
        self.clients.get(&id)
    }

    pub fn run(&mut self) {
        loop {
            match match self.event_hub.poll_event() {
                Event::Connect(client_id, responder) => self.handle_connection(client_id, responder),
                Event::Disconnect(client_id) => self.handle_disconnection(client_id),
                Event::Message(client_id, message) => self.handle_message(client_id, message),
            } {
                Ok(()) => {},
                Err(error) => println!("{:?}", error),
            }
        }
    }

    fn new_client(&mut self, client_id: u64, responder: Responder) -> &Client {
        let client = Client::new(ClientInfo::Ws {
            id: client_id,
            responder,
            discord_info: None,
        }, self.lobby.clone());
        self.clients.insert(client_id, client);
        self.clients.get(&client_id).unwrap()
    }

    fn handle_connection(&mut self, client_id: u64, responder: Responder) -> Result<(), ServerError> {
        // Debug
        println!("A client connected with id #{}", client_id);

        {
            // Initialize client
            let client = self.new_client(client_id, responder);

            // Get immutable access to room
            let room = client.room.borrow();

            // Send client greeting
            client.send(MessageResponseData::Greeting {
                id: client_id,
                next_mora: room.next_mora().clone(),
            }.into_message());

            // Sent recent message history
            client.send(MessageResponseData::History {
                words: room.get_history()?,
            }.into_message());
        }

        // Number of clients on Rc<RefCell<Room>> reference counter will be one more
        self.broadcast_player_count(&self.lobby);

        Ok(())
    }

    fn handle_disconnection(&mut self, client_id: u64) -> Result<(), ServerError> {
        // Debug
        println!("Client #{} disconnected.", client_id);

        // Remove client
        // At this point, client should be dropped
        let client = self.clients.remove(&client_id).unwrap();

        // Get room
        let room = client.room;

        // Number of clients on Rc<RefCell<Room>> reference counter will be one less
        self.broadcast_player_count(&room);

        Ok(())
    }

    fn for_client_in_room(&self, room: &Rc<RefCell<Room>>, mut closure: impl FnMut(&Client) -> ()) {
        for (_id, client) in self.clients.iter().filter(|(_id, client)| Rc::<RefCell<Room>>::ptr_eq(room, &client.room)) {
            closure(client);
        }
    }

    fn client_count_in_room(&self, room: &Rc<RefCell<Room>>) -> usize {
        let mut count: usize = 0;
        self.for_client_in_room(room, |_| count += 1);
        count
    }

    fn broadcast_player_count(&self, room: &Rc<RefCell<Room>>) {
        let response = MessageResponseData::PlayerCount { players: self.client_count_in_room(room) as u64 }.into_response();
        for (_id, client) in self.clients.iter() {
            client.send(response.to_message());
        }
    }

    fn announce_to_room(&self, room: &Rc<RefCell<Room>>, response_data: MessageResponseData) {
        let response = response_data.into_response();
        self.for_client_in_room(room, |client| {
            client.send(response.to_message());
        });
    }

    fn handle_message(&mut self, client_id: u64, message: Message) -> Result<(), ServerError> {
        // Ignore binary messages
        let message = match message {
            Message::Text(message) => {
                // Debug
                println!("Received a message from client #{}: {:?}", client_id, message);

                message
            }
            Message::Binary(message) => {
                // Debug
                println!("Received a binary message from client #{}: {:?}", client_id, message);

                return Ok(());
            }
        };

        let client = self.get_client(client_id).unwrap();

        match client.room.borrow_mut().handle_query(&message, client_id) {
            // Broadcast new words to all clients in same room
            Ok(response) => self.announce_to_room(&client.room, response),
            // Send errors to only this client
            Err(message) => {
                client.send(MessageResponseData::Error { message: message.to_string() }.into_message());
            }
        };
        Ok(())
    }
}
